#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "kohn_avi.h"

/*

This code is copyright 2008-2011 by Michael Kohn
This code falls under the LGPL license

http://www.mikekohn.net/
Michael Kohn <mike@mikekohn.net>

*/

int get_jpeg_dims(FILE *in, int *width, int *height)
{
int ch;

  while((ch=getc(in))!=EOF)
  {
    if (ch==0xff)
    {
      ch=getc(in);
      if (ch==0xc0)
      {
        ch=getc(in);
        ch=getc(in);
        ch=getc(in);
        ch=getc(in);
        *height=(ch<<8)+getc(in);
        ch=getc(in);
        *width=(ch<<8)+getc(in);
        return 0;
      }
    }
  }

  return -1;
}

int count_frames(char *filename_template)
{
FILE *in;
char filename[1024] = { 0 };
int t;

  printf("Scanning...\n");

  t=0;
  while(1)
  {
    sprintf(filename,filename_template,t);
    in=fopen(filename,"r");
    if (in==0) break;
    fclose(in);
    t++;
    if ((t%1000)==0) printf("%d frames so far...\n",t);
    if (strcmp(filename,filename_template)==0) break;
  }

  return t;
}

int main(int argc, char *argv[])
{
FILE *in;
int fd;
struct kavi_t *kavi;
char filename[1024];
unsigned char *buffer=0;
int buffer_len=0;
int fps,count,t,l,c,r;
int width,height;
struct stat file_stat;

  printf("\njpeg2avi routines - Copyright 2005-2011 by Michael Kohn\n");
  printf("http://www.mikekohn.net/\n");
  printf("Version Date: 2011-Jan-30\n\n");

  if (argc!=4)
  {
    printf("This program will take a set of numbered jpegs's and\n");
    printf("creates an avi/jpeg movie file.\n\n");
    printf("Usage: jpeg2avi <outfile.avi> <printf formatted filename> <frames per second>\n\n");
    printf("Example: jpeg2avi out.avi frames%%05d.jpeg 10\n\n");
    exit(0);
  }

  fps=atoi(argv[3]);
  if (fps<0)
  {
    printf("Illegal value for frames per second\n");
    exit(1);
  }

  count=count_frames(argv[2]);
  printf("Found %d frames\n",count);

  sprintf(filename,argv[2],0);
  in=fopen(filename,"rb");
  if (in==0)
  {
    printf("Cannot open jpeg %s\n",filename);
    exit(1);
  }

  if (get_jpeg_dims(in,&width,&height)!=0)
  {
    printf("Cannot find width/height of jpeg\n");
    exit(1);
  }

  fclose(in);

  printf("Video width=%d height=%d\n",width,height);

  kavi=kavi_open(argv[1],width,height,"MJPG",atoi(argv[3]),NULL);

  if (kavi==0)
  {
    printf("Couldn't open AVI for writing\n");
    exit(1);
  }

  for (t=0; t<count; t++)
  {
    sprintf(filename,argv[2],t);
    fd=open(filename,O_RDONLY);
    if (fd==-1) break;

    fstat(fd,&file_stat);
    l=file_stat.st_size;

    if (l>buffer_len)
    {
      if (buffer_len!=0) free(buffer);
      buffer_len=l;
      buffer=(unsigned char *)malloc(buffer_len);
    }

    c=0;
    while(c<l)
    {
      r=read(fd,buffer+c,l-c);

      c=c+r;
      if (r<=0) break;
    }
    close(fd);

    kavi_add_frame(kavi,buffer, l);

    if ((t%1000)==0) printf("%d frames so far...\n",t);
  }

  if (buffer_len!=0) free(buffer);

  kavi_close(kavi);

  return 0;
}


